mod config;
mod layers;
mod medium;
mod tun_worker;

use crate::config::IpOverXConfig;
use crate::tun_worker::tun_rx_worker::TunRxWorker;
use clap::Parser;
use log::LevelFilter;
use std::net::IpAddr;
use std::path::PathBuf;
use tun::Configuration;
use tun_worker::TunWorker;
use tun_worker::tun_tx_worker::TunTxWorker;

/// Send IP packets over anything*
#[derive(Parser, Debug)]
#[command(version, about, long_about = None)]
struct Args {
    /// Net Mask to use
    #[arg(long, default_value = "255.255.255.0")]
    net_mask: IpAddr,
    /// Config file path
    cfg_path: PathBuf,
    /// Local TUN device IP
    local_addr: IpAddr,
    /// Remote TUN device IP
    remote_addr: IpAddr,
}

fn main() -> Result<(), Box<dyn std::error::Error + Send + Sync + 'static>> {
    env_logger::builder()
        .filter_level(LevelFilter::Info)
        .parse_default_env()
        .init();

    let args = Args::parse();

    let config = IpOverXConfig::new(&args.cfg_path)?;

    log::info!("Creating TUN device...");
    let mut tun_config = Configuration::default();
    tun_config
        .address(args.local_addr)
        .netmask(args.net_mask)
        .destination(args.remote_addr)
        .up();

    #[cfg(target_os = "linux")]
    tun_config.platform_config(|config| {
        // requiring root privilege to acquire complete functions
        config.ensure_root_privileges(true);
    });

    log::info!(
        "Opening tun device @ {} connecting to {}...",
        args.local_addr,
        args.remote_addr
    );
    let dev = tun::create(&tun_config)?;

    let (tx, rx) = dev.split();

    log::info!("Opening Medium...");
    let mut medium = config.medium.create_medium();

    medium.init().unwrap();

    let (rx_medium, tx_medium) = medium.split();

    let mut tx_worker = TunTxWorker::new(tx, tx_medium);
    let mut rx_worker = TunRxWorker::new(rx, rx_medium);

    log::info!("Adding layers...");
    for layer in &config.layers {
        let layer = layer.get_layer();
        log::info!("Adding '{}' layer", layer.name());
        let (rx_layer, tx_layer) = layer.split();
        tx_worker.add_layer(tx_layer);
        rx_worker.add_layer(rx_layer);
    }

    log::info!("Starting worker threads...");
    let tx_thread = std::thread::spawn(move || tx_worker.worker_thread());
    let rx_thread = std::thread::spawn(move || rx_worker.worker_thread());

    log::info!("Running, waiting for exit...");
    tx_thread.join().unwrap();
    rx_thread.join().unwrap();

    Ok(())
}
