use crate::medium::{Medium, ReadMedium, WriteMedium};
use log::info;
use std::io::{ErrorKind, Read, Write};
use std::net::{SocketAddr, UdpSocket};

pub struct UdpMedium {
    source_addr: SocketAddr,
    dest_addr: SocketAddr,
    udp_socket: Option<UdpSocket>,
}

impl UdpMedium {
    pub fn new(source_addr: SocketAddr, dest_addr: SocketAddr) -> Self {
        Self {
            source_addr,
            dest_addr,
            udp_socket: None,
        }
    }
}

impl Medium for UdpMedium {
    fn split(&self) -> (Box<dyn ReadMedium>, Box<dyn WriteMedium>) {
        let rx_medium = Self {
            source_addr: self.source_addr,
            dest_addr: self.dest_addr,
            udp_socket: Some(self.udp_socket.as_ref().unwrap().try_clone().unwrap()),
        };

        let tx_medium = Self {
            source_addr: self.source_addr,
            dest_addr: self.dest_addr,
            udp_socket: Some(self.udp_socket.as_ref().unwrap().try_clone().unwrap()),
        };

        (Box::new(rx_medium), Box::new(tx_medium))
    }

    fn name(&self) -> String {
        "Udp".to_string()
    }

    fn init(&mut self) -> Result<(), std::io::Error> {
        info!(
            "Initializing Udp Medium @ {} connected to {}",
            self.source_addr, self.dest_addr
        );
        self.udp_socket = Some(UdpSocket::bind(self.source_addr)?);

        Ok(())
    }
}

impl Write for UdpMedium {
    fn write(&mut self, buf: &[u8]) -> std::io::Result<usize> {
        if let Some(socket) = &mut self.udp_socket {
            socket.send_to(buf, self.dest_addr)
        } else {
            Err(std::io::Error::from(ErrorKind::NotConnected))
        }
    }

    fn flush(&mut self) -> std::io::Result<()> {
        // Nothing needed
        Ok(())
    }
}

impl WriteMedium for UdpMedium {}

impl Read for UdpMedium {
    fn read(&mut self, buf: &mut [u8]) -> std::io::Result<usize> {
        if let Some(socket) = &mut self.udp_socket {
            socket.recv(buf)
        } else {
            Err(std::io::Error::from(ErrorKind::NotConnected))
        }
    }
}

impl ReadMedium for UdpMedium {}
